<?php

namespace App\Controller;

use App\Exception\ItemException as ItemException;
use App\Exception\IntegrationException as IntegrationException;
use App\Exception\ParamsException as ParamsException;
use App\Exception\SqlException as SqlException;
use App\Exception\CancelException as CancelException;
use App\Exception\NetworkException as NetworkException;
use App\Exception\GeneralErrorException as GeneralErrorException;
use App\Exception\MenuException as MenuException;


use Psr\Http\Message\ResponseInterface as Response;
use Psr\Http\Message\ServerRequestInterface as Request;

class WebService {

    const FAILOVER_EMAIL = "daniel.fonseca@teknisa.com";
    //const FAILOVER_EMAIL = "daniel.santos@teknisa.com";

    private $connection;
    private $branchService;
    private $consumerService;
    private $integrationService;
    private $orderService;
    private $paymentService;
    private $productService;
    private $util;
    

    
    private $mailList;
    
    const API_ANT_BASE_URL  = "https://api-parceiros.anota.ai/";
    const GEOCODE_URL       = "https://maps.google.com/maps/api/geocode/json?key=AIzaSyBfblcHNG4xb8lcwzD5sUvQ-dXIuY3QbNs=";

    public function __construct(\Doctrine\DBAL\Connection $connection,
                                \App\Service\BranchService $branchService,
                                \App\Service\ConsumerService $consumerService,
                                \App\Service\IntegrationService $integrationService,
                                \App\Service\OrderService $orderService, 
                                \App\Service\PaymentService $paymentService,
                                \App\Service\ProductService $productService,
                                \App\Util\Util $util)
                                {

        $this->connection         = $connection;
        $this->branchService      = $branchService;
        $this->consumerService    = $consumerService;
        $this->integrationService = $integrationService;
        $this->orderService       = $orderService;
        $this->paymentService     = $paymentService;
        $this->productService     = $productService;
        $this->util               = $util;
    }

    public function showVersion(Request $request, Response $response)
    {
        $sql        = "SELECT CDFILIAL FROM FILIAL";
        $result     = $this->connection->fetchAll($sql);

        $versionInfo = $this->getVersionInfo();

        return $response->withJson([
            'version'     => $versionInfo['version'],
            'revision'    => $versionInfo['revision'],
            'date'        => $versionInfo['date'],
            'branchs'     => count($result)
            //'database'    => ''//$this->connection->getParams()
        ]);
    }
    
    private function getVersionInfo()
    {
        $file = file_get_contents(__DIR__ . "/../../version.json");
        return json_decode($file, true);
    }

    /**
    * Função de entrada dos pedidos provenientes da Anota Aí, os pedidos chegam pelo WebHook respectivo
    *
    * @param Request $request
    * @param Response $response
    * @throws ParamsException Exceção de parâmetros inválidos ou faltantes.
    * @throws ItemException Exceção de parâmetros inválidos nos itens do pedido.
    * @throws SqlException Exceção de erros SQL.
    * @throws Exception Exceções gerais.
    * @return void
    */
    public function runWebService(Request $request, Response $response)
    {
        $params = $request->getParsedBody();

        // $this->debugLog($params);

        if ($params["auth"] != "81743A8E8D41AF82A4807097CDF4BEC2DDDB9E6334BECCFED4D46FC783A8CA2E") {
            $response->withStatus(400);
            return 0;
        }
            
        $errorMessage = "";
        $rawOrder     = "";
        $writeEmail   = true;
        
        $configuration = $this->integrationService->getConfiguration();
        if(empty($configuration)) {
            $response->withStatus(200);
            return 0;
        }

        $ufTimezone = (isset($configuration['IDUFTIMEZONE']) && !empty($configuration['IDUFTIMEZONE'])) ? $configuration['IDUFTIMEZONE'] : "SP";
        
        date_default_timezone_set($this->util->getCurrentTimezone(strtoupper($ufTimezone)));
        
        if($configuration['IDHABINTE'] == 'N') 
            return $response->withStatus(200);
        
        $this->mailList = $this->util->getMailList();
        $this->connection->beginTransaction();
        
        $dateLog = date("d/m/Y H:i:s");
        
        try {
            $this->mailList = $this->util->getMailList();
            $order          = $params["anotaAiData"];
            
            $debug          = isset($params['debug']) ? $params['debug'] : false;
            $merchantId     = $order["merchant"]["id"];
            $orderId        = $order["id"];
            $shortId        = $order["shortReference"];
            
            $rawOrder       = $order;
            $deliveryParams = $this->orderService->getDeliveryMethodParams($merchantId);
    
            if(empty($deliveryParams['IDTOKENINTGERAL']))
                throw new GeneralErrorException("Token geral não encontrado, associado ao id ".$merchantId." do estabelecimento.");
            else{
                $this->util->setAccessToken($deliveryParams['IDTOKENINTGERAL']);
            }    

            $deliveryParams['insertChange'] = ($deliveryParams['IDRETTXENTR'] == "N");

            $order["cdFilial"] = $deliveryParams["CDFILIAL"];
            
            $fileName = ((!empty($order["cdFilial"])) && (!empty($shortId))) ? "{$order['cdFilial']}-{$shortId}-{$this->util->getCurrentDateTime()}.json" : ("INDEFINIDO-" . $this->util->getCurrentDateTime()) . ".json";
            
            $this->util->writeLog('order', $fileName, $order);
            $this->util->writeLog("execution", "", $dateLog. ": [  Pedido Recebido  ] #" . $shortId . " - Filial: " . $order['cdFilial'] . "\n");
            
            $integrationDetails = $this->processOrder($order);

            $branchId = $integrationDetails["CDFILIAL"];
            $sellId  = $integrationDetails["NRVENDAREST"];
            $this->orderService->validateOrder($branchId, $sellId);
            $this->util->writeLog('order_processed', $fileName, $order);
            
            if(!$debug) {
              $acceptOrder = $this->acceptOrder($orderId);
              $this->util->writeLog('accepted', $fileName, $acceptOrder);
            }
            $this->connection->commit();
            
            $this->util->writeLog("execution", "", $dateLog. ": [ Pedido Integrado  ] #" . $shortId . " -> Sucesso.\n");
            
            $this->itsAlive();
            
            return $response->withStatus(200);

        } catch(ItemException $e) {
            $errorMessage = date("d-m-Y H:i:s") . " - " . $e->getMessage();
            if(empty($fileName)) $fileName = "INDEFINIDO-{$this->util->getCurrentDateTime()}.json";
            
            $this->util->writeLog("order_error", $fileName, $errorMessage);
            $this->connection->rollback();
            if(!$debug){
                $responseReject = $this->rejectOrder($orderId);    
                $this->util->writeLog('rejected', $fileName, $responseReject);
            }
        } catch(ParamsException $e) {
            $errorMessage = date("d-m-Y H:i:s") . " - " . $e->getMessage();
            if(empty($fileName)) $fileName = "INDEFINIDO-{$this->util->getCurrentDateTime()}.json";
            
            $this->util->writeLog("order_error", $fileName, $errorMessage);
            $this->util->writeLog("execution", "", $dateLog . ": [Erro Parametrizacao] #" . $shortId . " -> ". $e->getMessage() . "\n");
            $this->connection->rollback();
            if(!$debug){
                $responseReject = $this->rejectOrder($orderId);    
                $this->util->writeLog('rejected', $fileName, $responseReject);
            }
        } catch(SqlException $e) {
            $errorMessage = date("d-m-Y H:i:s") . " - " . $e->getMessage();
            if(empty($fileName)) $fileName = "INDEFINIDO-{$this->util->getCurrentDateTime()}.json";
            
            $this->util->writeLog("order_error", $fileName, $errorMessage);
            $this->util->writeLog("execution", "", $dateLog . ": [     Erro  SQL     ] #" . $shortId . " -> ". $e->getMessage() . "\n");
            $this->connection->rollback();
            if(!$debug){
                $responseReject = $this->rejectOrder($orderId);    
                $this->util->writeLog('rejected', $fileName, $responseReject);
            }
        } catch(NetworkException $e) {
            $errorMessage = date("d-m-Y H:i:s") . " - " . $e->getMessage();
            if(empty($fileName)) $fileName = "INDEFINIDO-{$this->util->getCurrentDateTime()}.json";
            
            $this->util->writeLog("network", $fileName, $errorMessage);
            $this->util->writeLog("execution", "", $dateLog . ": [   Erro de Rede    ] -> {$e->getMessage()}\n");
            $this->connection->rollback();
            if(!$debug){
                $responseReject = $this->rejectOrder($orderId);    
                $this->util->writeLog('rejected', $fileName, $responseReject);
            }
        } catch(IntegrationException $e) {
            $writeEmail = false;
            $errorMessage = date("d-m-Y H:i:s") . " - " . $e->getMessage();
            if(empty($fileName)) $fileName = "INDEFINIDO-{$this->util->getCurrentDateTime()}.json";
            
            $this->util->writeLog("order_error", $fileName, $errorMessage);
            $this->connection->rollback();
            if(!$debug){
                $responseReject = $this->rejectOrder($orderId);    
                $this->util->writeLog('rejected', $fileName, $responseReject);
            }
        } catch(CancelException $e) {
            $errorMessage = date("d-m-Y H:i:s") . " - " . $e->getMessage();
            if(empty($fileName)) $fileName = "INDEFINIDO-{$this->util->getCurrentDateTime()}.json";
            
            $this->util->writeLog("cancelled_error", $fileName, $errorMessage);
            $this->connection->rollback();
        } catch(GeneralErrorException $e) {
            $errorMessage = date("d-m-Y H:i:s") . " - " . $e->getMessage();
            if(empty($fileName)) $fileName = "INDEFINIDO-{$this->util->getCurrentDateTime()}.json";
            
            $this->util->writeLog("error", $fileName, $errorMessage);
            $this->connection->rollback();
            // desligar projeto            
        } catch(\Exception $e) {
            $errorMessage = date("d-m-Y H:i:s") . " - " . $e->getMessage();
            if(empty($fileName)) $fileName = "INDEFINIDO-{$this->util->getCurrentDateTime()}.json";
            
            $this->util->writeLog("error", $fileName, $errorMessage);
            $this->connection->rollback();
            if(!$debug){
                $responseReject = $this->rejectOrder($orderId);    
                $this->util->writeLog('rejected', $fileName, $responseReject);
            }
        } 
        if (!empty($errorMessage)) {
            $cdFilial    = $this->getBranch($merchantId);
            $orderNumber = (isset($orderId) ? 
                        ("$cdFilial-{$orderId}") : ("INDEFINIDO-" . date("H-i-s"))); 
            
            /* ----- DEPRECIADO            
            $this->util->writeLog('order_error', date('H-i-s') . '.txt', $errorMessage);
            */
            
            if($writeEmail) {
                global $CLIENT_NAME;
                $clientName = $CLIENT_NAME;

                $mailBody = file_get_contents(__DIR__ . "/../../templates/defaultMail.html");

                if($debug) $this->mailList = null; 

                $this->util->sendMail((!empty($this->mailList) ? $this->mailList : [self::FAILOVER_EMAIL]), [], 'Integração Teknisa x Anota Aí', $mailBody, [
                    ":INTEGRACAO"    => 'Teknisa x Anota Aí',
                    ":NUMERO_PEDIDO" => $orderNumber,
                    ":NOME_CLIENTE"  => $clientName,
                    ":NOME_SERVIDOR" => "",
                    ":ERRO_MSG"      => $errorMessage,
                    ":JSON_PEDIDO"   => json_encode($rawOrder, JSON_PRETTY_PRINT | JSON_UNESCAPED_UNICODE),
                    ":GERADO_EM"     => date("d/m/Y H:i:s")
                ]);
            }
        }

        return $response->withStatus(500);
    }

    // private function debugLog($params)
    // {
    //     $folder      = "/home/developer/workfolder/logs/anotaAi/debugLog";
    //     $fileName = round(microtime(true)*1000).'_order';
    //     $currentFile = "$folder/$fileName";
    //     if(!file_exists($folder)) {
    //         mkdir($folder, 0775, true);
    //     }
    //     file_put_contents($currentFile, json_encode($params, JSON_PRETTY_PRINT | JSON_UNESCAPED_UNICODE));
    // }

    private function processOrder($order)
    {
        $generalParams = $this->integrationService->generalParameters($order['cdFilial']);
        
        // TOGO - Insert branch address to order
        if($order["type"] == "TAKE") {
            $branchInfoAddress                           = $this->branchService->getBranchAddress($order['cdFilial']);
            $order["deliveryAddress"]["state"]           = $branchInfoAddress["SGESTADO"];
            $order["deliveryAddress"]["city"]            = $branchInfoAddress["CDMUNICIPIO"];
            $order["deliveryAddress"]["neighborhood"]    = $branchInfoAddress["NMBAIRFILI"];
            $order["deliveryAddress"]["streetNumber"]    = preg_replace("/[^0-9]/", "", $branchInfoAddress["DSENDEFILI"]);
            $order["deliveryAddress"]["streetName"]      = str_replace($order["deliveryAddress"]["streetNumber"], "", $branchInfoAddress["DSENDEFILI"]);
            $order["deliveryAddress"]["postalCode"]      = $branchInfoAddress["NRCEPFILI"];
            $order["deliveryAddress"]["complement"]      = "";
            $order["deliveryAddress"]["reference"]       = "";
            $order["deliveryAddress"]["lat"]             = "";
            $order["deliveryAddress"]["lng"]             = "";
        }

        $consumer = $this->consumerService->processConsumer($generalParams, $order);
        return $this->integrationService->openOrderBill($generalParams, $order, $consumer, $order["items"]);
    }

    private function getBranch($token)
    {
        $branch = $this->orderService->getBranch($token);

        if(empty($branch)) 
            throw new ParamsException("Nao foi encontrado filial associada ao token: $token.");
            
        return $branch;
    }

    private function saveVendarest($cdFilial, $cdloja, $nrmesa, $cdvendedor, $cdoperador, $cdcliente, $cdconsumidor, $sgestado)
    {
        return $this->orderService->saveVendarest($cdFilial, $cdloja, $nrmesa, $cdvendedor, $cdoperador, $cdcliente, $cdconsumidor, $sgestado);
    }

    private function saveComandaven($cdFilial, $nrvendarest, $nrcomandaext, $cdloja, $idstcomanda, $deliveryFee, $nrpedmobile, $consumerData, $dsObsComanda, $consumerIdentifier, $dlvMethod)
    {
        return $this->orderService->saveComandaven($cdFilial, $nrvendarest, $nrcomandaext, $cdloja, $idstcomanda, $deliveryFee, $nrpedmobile, $consumerData, $dsObsComanda, $consumerIdentifier, $dlvMethod);
    }
    
    private function saveMovcaixadlv($deliveryParams, $order, $cdFilial, $nrvendarest, $charges, $orderPaidInCash, $cdcliente, $cdconsumidor, $payments)
    {
        foreach($payments as $payment){
            if( $payment["code"] == "card" && empty($payment["cardSelected"]) ){
                $paymentCode = $this->paymentService->getPaymentCode();
            }
            elseif($payment["code"] == "card" && !empty($payment["cardSelected"])){
                $paymentCode = $this->paymentService->getReceiptType($cdFilial, "ANT", $payment["externalId"]);
            }
            elseif ($payment['code'] ==  'money'){
                $paymentCode = $this->paymentService->getMoneyReceiptCode();
            }    
            $this->paymentService->saveMovcaixadlv($deliveryParams, $order, $cdFilial, $nrvendarest, $charges, $paymentCode, $cdcliente, $cdconsumidor);
        }
    }

    private function acceptOrder($orderId)
    {
        
        $generalToken   = $this->util->getAccessToken();
        // $url         = self::API_ANT_BASE_URL."v1/order/order/confirm/?auth_token=$authToken&order_id=$orderId";
        $url         = self::API_ANT_BASE_URL."partnerauth/order/accept/$orderId";
        $body        = json_encode([]);
        
        $header = [
            "Content-type: application/json",
            "Authorization: " . $generalToken
        ];
    
        $response = $this->util->executeWebService($url, "POST", $header, $body);

        return json_decode($response, true);
    }

    private function rejectOrder($orderId)
    {
        $cancellationId = 922;
        // $reason         = "NAO FOI POSSIVEL REALIZAR INTEGRACAO";
        $generalToken      = $this->util->getAccessToken();
        // $url            = self::API_ANT_BASE_URL."v1/order/order/cancel/" . urlencode("?auth_token=$authToken&order_id=$orderId&reason_id=$cancellationId&reason=$reason");
        $url         = self::API_ANT_BASE_URL."partnerauth/order/cancel/$orderId";
        $body           = json_encode(["justification"=>"NAO FOI POSSIVEL REALIZAR INTEGRACAO"]);
        
        $header = [
            "Content-type: application/json",
            "Authorization: " . $generalToken
        ];
    
        $response = $this->util->executeWebService($url, "POST", $header, $body);

        return json_decode($response, true);
    }

    public function cancelOrder(Request $request, Response $response)
    {
        try {

            
            $body = $request->getParsedBody();
            $orderId = $body['anotaAiData']['id'];

            $this->debugLog($body);

            $this->orderService->cancelOrder($orderId);
            

            $order = $this->getOrderDetailsById($orderId);
            
            $fileName = ((!empty($order["CDFILIAL"])) && (!empty($orderId))) ? "{$order['CDFILIAL']}-{$orderId}-{$this->util->getCurrentDateTime()}.json" : ("INDEFINIDO-" . $this->util->getCurrentDateTime()) . ".json";
            // $fileName     = $order["CDFILIAL"] . "-" . $order["NRCOMANDAEXT"] . "-". date("c") . ".json";

            $this->util->writeLog('cancelled', $fileName, $body);
        } catch(CancelException $e) {
            $errorMessage = $e->getMessage();
        } catch(\Exception $e) {
            $errorMessage = $e->getMessage();
        } ;
        if (!empty($errorMessage)) {
            $this->util->writeLog('cancelled_error', date("d-m-Y H:i:s") . '.txt', $errorMessage);
            return $response->withStatus(500);
        }

        return $response->withStatus(200);
    }
    
    public function getOrderDetailsById($orderId)
    {
        return $this->orderService->getOrderDetailsById($orderId);
    }
    
    public function itsAlive()
    {
        $executionLogFilePath  = $this->util->getExecutionLog();
        $executionLogTimeStamp = filemtime($executionLogFilePath);
        if((time() - $executionLogTimeStamp) >= 3600) {
            $this->util->writeLog("execution", "", date("d/m/Y H:i:s") . ": [    Integridade    ] Verificacao de integridade.\n");
        }
    }
    
}