<?php

namespace App\Service;

use App\Exception;
use App\Exception\IntegrationException as IntegrationException;
use App\Exception\ParamsException as ParamsException;
use App\Exception\NetworkException as NetworkException;
use App\Exception\SqlException as SqlException;


class IntegrationService {
    
    private $util;
    private $productService;
    private $integrationFactory;
    private $branchFactory;
    private $orderFactory;
    private $paymentFactory;

    public function __construct (\App\Util\Util $util, 
                                 \App\Service\ProductService $productService, 
                                 \App\Factory\IntegrationFactory $integrationFactory,
                                 \App\Factory\BranchFactory $branchFactory, 
                                 \App\Factory\OrderFactory $orderFactory, 
                                 \App\Factory\PaymentFactory $paymentFactory)
                                 {
        $this->util               = $util;
        $this->productService     = $productService;
        $this->integrationFactory = $integrationFactory;
        $this->branchFactory      = $branchFactory;
        $this->orderFactory       = $orderFactory;
        $this->paymentFactory     = $paymentFactory;
    }
    
    public function getConfiguration()
    {
        return $this->integrationFactory->getConfiguration();
    }
    
    public function generalParameters($cdFilial)
    {
        $vendor        = $this->branchFactory->getStandartVendor($cdFilial);
        $cashier       = $this->branchFactory->getStandartCashier($cdFilial, $vendor);
        $store         = $this->branchFactory->getStandartStore($cdFilial, $vendor);
        $idControlProd = $this->branchFactory->getProductControlId($cdFilial, $store["CDLOJA"]);
        $state         = $this->branchFactory->getBranchState($cdFilial);
        
        return [
            "CDFILIAL"     => $cdFilial,
            "CDVENDEDOR"   => $vendor,
            "CDLOJA"       => $store["CDLOJA"],
            "CDCAIXA"      => $cashier["CDCAIXA"],
            "CDOPERADOR"   => $cashier["CDOPERADOR"],
            "IDCONTROPROD" => $idControlProd,
            "SGESTADO"     => $state,
            "IMPAUTO"      => (!empty($store["IMPAUTO"]) && $store["IMPAUTO"] != 'N') ? true : false
        ];
    }
    
    public function openOrderBill($generalParams, $order, $consumer, $products)
    {
        if(empty($generalParams["CDOPERADOR"])) {
            throw new Exception\ParamsException("
                Não existe um operador padrão atrelado ao vendedor padrão {$generalParams['CDVENDEDOR']} da filial {$generalParams['CDFILIAL']}.
                Para atrelar o operador padrão ao vendedor padrão desta filial, verifique a seção 'Problemas Frequentes', no manual da integração.");
        } 
        
        $operator = $this->orderFactory->validateOperator($generalParams);
        
        if(empty($operator)) {
            throw new Exception\ParamsException("
                O operador {$generalParams['CDOPERADOR']} não está associado a filial {$generalParams['CDFILIAL']}.<br>
                Para atrelar o operador a esta filial, verifique a seção 'Problemas Frequentes', no manual da integração.");
        }

        if (!$this->validateDuplication($order["shortReference"], $order["cdFilial"])){
            throw new IntegrationException("Pedido ". $order["shortReference"] .", para a filial ". $order["cdFilial"] . " já integrado pelo sistema.");
        }
        
        $sale = $this->util->newCode("VENDAREST".$generalParams["CDFILIAL"], 10);
        $comd = $this->util->newCode("COMANDAVEN".$generalParams["CDFILIAL"], 10);
        
        $order["NRVENDAREST"] = $sale;
        $order["NRCOMANDA"]   = $comd;
        $order["IDSTCOMANDA"] = $generalParams["IDCONTROPROD"] === "S" ? "1" : "3";
        
        if(isset($order["schedule_order"]) && !empty($order["schedule_order"])) {
            $utcTimezone   = new \DateTimeZone("UTC");
            $orderDateTime = new \DateTime($order["schedule_order"]["date"], $utcTimezone);
            $storeTimezone = new \DateTimeZone($this->util->getCurrentTimezone(strtoupper($generalParams["SGESTADO"])));
            
            $orderDateTime->setTimezone($storeTimezone);
            
            $order["DTHRAGENDADA"] = date_format($orderDateTime, "d/m/Y H:i:s");
            $order["IDSTCOMANDA"]  = "A";
            $order["IDSINCAGENDA"] = "N";
        } else {
            $order["DTHRAGENDADA"] = null;
            $order["IDSTCOMANDA"]  = $generalParams["IDCONTROPROD"] === "S" ? "1" : "3";
            $order["IDSINCAGENDA"] = "S";
        }
        
        // Observação de pagamento online
        $generalParams['DSOBSCOMANDA'] = "";

        if(!empty($order["payments"])) {
            $paymentName = $order["payments"][0]["name"];
            $str = "{$paymentName}";
            $str .= $order["payments"][0]["changeFor"] > "0.00" ? " (VALOR: {$order['payments'][0]['changeFor']}); ": " (VALOR: {$order['payments'][0]['value']}); ";
            $generalParams['DSOBSCOMANDA'] .= $str;
        }
        
        if(isset($order["observation"]) && !empty($order["observation"])){
            $generalParams['DSOBSCOMANDA'] .= $order["observation"]; 
        }
        
         /* Fechamento da comanda */
        $this->orderFactory->insertSale($sale, $generalParams, $consumer);
        $this->orderFactory->insertSaleCommand($generalParams, $order, $consumer);
            
        /* Fechamento de itens */
        
        $total = $order["total"];

        //MOVCAIXADLV
        $change = "0.00";
        $moneyCode = $this->paymentFactory->getMoneyReceiptCode();
        
        $paymentMethod        = $order["payments"][0]["code"];
        $order["CDTIPORECE"]  = $this->paymentFactory->getReceiptType($paymentMethod, $generalParams["CDFILIAL"]);
        $order["NRSEQMOVDLV"] = $this->util->newCode("MOVCAIXADLV".$generalParams["CDFILIAL"].$sale, 10);
        
        $paymentValue = $total;

        if(($order["CDTIPORECE"] == $moneyCode) && $order["payments"][0]["changeFor"] != null) {
            $change = bcsub($order["payments"][0]["changeFor"], $order["payments"][0]["value"], 2);
            $paymentValue = $order["payments"][0]["changeFor"];
        }
        
        $this->orderFactory->insertCashMoviment($order, $paymentValue, $consumer);

        if($change > "0.00") {
            $movCaixa = $this->util->newCode("MOVCAIXADLV".$generalParams["CDFILIAL"].$sale, 10);
            $this->orderFactory->insertChange($order, $change, $generalParams, $consumer, $movCaixa, $moneyCode);
        }
        
        $generalParams["NRSEQPRODCUP"] = $this->util->newCode("ITCOMANDAVEN".$order["cdFilial"], 10);
        
        $products     = [];
        $nrSeqProdCom = 0;
        
        $ITCOMANDAVEN_SEQ = 0;

        foreach ($order["items"] as $item) {
            
            if(empty($item['externalId']) && ($item['price'] != null && $item['price'] != "0.00")){
                throw new Exception\ParamsException("
                        O produto {$item['name']} possui preço e não possui código externo.<br>
                        Para realizar a correção, entre em contato com a equipe de suporte do Delivery Direto.");
            }
            
            if(!empty($item['externalId']) && ($item['price'] == null || $item['price'] == "0.00")) {
                    throw new Exception\ParamsException("
                        O produto {$item['name']} possui preço inferior a R$ 0,01 centavo.<br>
                        Código do produto: {$item['externalId']}.<br>
                        Para realizar a correção, entre em contato com a equipe de suporte do Delivery Direto.");
            }
            
            $processedProduct = $this->productService->processProduct($generalParams, $item, $order, $nrSeqProdCom, $ITCOMANDAVEN_SEQ);
            
            if (is_array($processedProduct) && !isset($processedProduct["CDPRODUTO"])) {
                $products = array_merge($products, $processedProduct);
            } else {
                array_push($products, $processedProduct);
            }
        }

        if(empty($products)) {
            throw new \Exception("Ocorreu uma falha na validação de produtos.", 187);
        }
        
        $totalDiscount = 0;

        foreach($order["discounts"] as $discount) {
            if($discount["target"] == "CART") {
                $totalDiscount += $discount["amount"];
            } else if ($discount["target"] == "DELIVERY_FEE") {
                $order["deliveryFee"] -= $discount["amount"];
                if($discount["amount"] > $order["deliveryFee"]) {
                    $order["deliveryFee"] = 0;
                }
            } else {
                throw new ParamsException ("Desconto não reconhecido.");
            }
        }
        
        if($totalDiscount != 0) {
            $this->shareDiscount($products, $totalDiscount);
        }

        foreach ($products as $product) {
            if(empty($product['VRPRECCOMVEN']) || $product['VRPRECCOMVEN'] < 0) 
                throw new \Exception("O produto ({$product['CDPRODUTO']}) possui preço inferior a 0.01 centavo.");
            
            if($product['VRDESCCOMVEN'] >= $product['VRPRECCOMVEN']) 
                throw new \Exception("O produto ({$product['CDPRODUTO']}) possui DESCONTO superior ao VALOR: Desconto -> {$product['VRDESCCOMVEN']} Valor -> {$product['VRPRECCOMVEN']}", 192);
            
            $this->productService->insertProduct(array_merge($product, $generalParams));
            
            //Observações da OBSITCOMANDAVEN
            if(!empty($product['ITEM_OBSERVATION'])) {
                foreach($product['ITEM_OBSERVATION'] as $itemObservation) {
                    $this->productService->insertObservation($itemObservation);
                }
            }
        }
        
        return [
            'CDFILIAL'      => $generalParams["CDFILIAL"],
            'NRCOMANDA'     => $order["NRCOMANDA"],
            'NRCOMANDAEXT'  => $order["shortReference"],
            'NRVENDAREST'   => $order["NRVENDAREST"],
            'IMPAUTO'       => $generalParams["IMPAUTO"]
        ];
    }
    
    public function validateDuplication($order, $branch)
    {
        return $this->orderFactory->validateDuplication($order, $branch);
    }
    
    public function shareDiscount(&$products, $initialDiscount)
    {
        $idxToApplyDiscount   = [];
        $totalDiscount        = $initialDiscount;
        $discountLeft         = $initialDiscount;
        $productsTotalValue   = "0.00";
        $discountGivenShare   = "0.00";
        $maxValueItem         = "0.00";
        $maxValueIdx          = -1;
        
        foreach($products as &$product) {

            if(!isset($product['VRDESCCOMVEN']) || $product['VRDESCCOMVEN'] == null){
                $product['VRDESCCOMVEN'] = 0;
            } 
            $product['discountShare'] = "0.00";
        }
        
        if(isset($product))
            unset($product);

        foreach($products as $i => $product) {
            if((bccomp($product["VRPRECCOMVEN"], "0.01", 2) == 1)) { 
                $productsTotalValue = bcadd($productsTotalValue, $product["VRPRECCOMVEN"], 2);
                array_push($idxToApplyDiscount, $i);
            }
        }

        foreach($idxToApplyDiscount as $idxWithDiscount) {
            $discountShare = bcdiv((bcmul($products[$idxWithDiscount]["VRPRECCOMVEN"], "100.00", 2)), $productsTotalValue, 2);
            $products[$idxWithDiscount]["discountShare"] = $discountShare;
            $discountGivenShare = bcadd($discountGivenShare, $discountShare, 2);
        }

        foreach($products as $i => $product) {
            if(bccomp($product['discountShare'], $maxValueItem, 2) == 1) {
                $maxValueIdx  = $i;
                $maxValueItem = $product['discountShare'];
            }            
        }

        if($maxValueIdx != -1) {
            $products[$maxValueIdx]["discountShare"] = bcadd($products[$maxValueIdx]["discountShare"], bcsub("100.00", $discountGivenShare, 2), 2);
        }
    
        foreach($idxToApplyDiscount as $idxWithDiscount) {
            $itemDiscount = bcdiv((bcmul($products[$idxWithDiscount]["discountShare"], $totalDiscount, 2)), "100.00", 2);
            $products[$idxWithDiscount]["VRDESCCOMVEN"] += $itemDiscount;

            $discountLeft = bcsub($discountLeft, $itemDiscount, 2);
        }

        if($maxValueIdx != -1) {
            $products[$maxValueIdx]["VRDESCCOMVEN"] = bcadd($products[$maxValueIdx]["VRDESCCOMVEN"], $discountLeft, 2);
            $discountLeft = "0.00";
        }
        
        $leftDiscount = "0.00";
        foreach($products as &$product) {
            if( (bccomp($product['VRDESCCOMVEN'], $product['VRPRECCOMVEN'], 2) == 1) || (bccomp($product['VRDESCCOMVEN'], $product['VRPRECCOMVEN'], 2) == 0) )  {
                $leftDiscount = bcadd($leftDiscount, bcsub($product['VRDESCCOMVEN'], $product['VRPRECCOMVEN'], 2), 2);
                $leftDiscount = bcadd($leftDiscount, "0.01", 2);
                
                $product['VRDESCCOMVEN'] = bcsub($product['VRDESCCOMVEN'], $leftDiscount, 2);
            }
        }
        
        if(isset($product))
            unset($product);

        foreach($products as &$product){
            $maxAddDisc = (bcsub(bcsub($product['VRPRECCOMVEN'], "0.01", 2), $product['VRDESCCOMVEN'], 2));
            if($maxAddDisc >= $leftDiscount) {
                $addDisc = $leftDiscount;
            } else{
                $addDisc = $maxAddDisc;
            }
            
            $product['VRDESCCOMVEN'] = bcadd($product['VRDESCCOMVEN'], $addDisc, 2);
            $leftDiscount = bcsub($leftDiscount, $addDisc, 2);
        }

        if(isset($product))
            unset($product);

        return $products;
    }
}